classdef RWCoseg < QPmv & BoxConstraints
    % A minimization problem of the form
    %
    %   min   x_1' * L_1 * x_1 + x_2' * L_1 * x_2 
    %         +  lambda * |H_1 * x_1 - H_2 * x_2|_2^2
    %
    %   s.t.  0 <= x <= 1
    %
    % The objective can also be written as a QP
    %
    %   1/2 * x' * Q * x
    %
    % Although crucially we don't explicitly represent H' * H in this
    % class, as the product has a large number of nonzeros.  Rather,
    % the matrix-vector product is calculated with H' * (H * x).
    %
    %
    
    properties
        L
        lambda
        H
        hbar
        seeds
        
        nimages
        start
        stop
        npixels
    end
    
    methods
        function x = pack(obj, cells)
            nimages = obj.nimages;
            n = obj.stop(nimages);
            
            x = zeros(n, 1);
            for i=1:nimages
                x(obj.start(i):obj.stop(i)) = cells{i};
            end
        end
        
        function cells = unpack(obj, x)
            nimages = obj.nimages;
            
            cells = cell(nimages, 1);
            for i=1:nimages
                cells{i} = x(obj.start(i):obj.stop(i));
            end
        end
        
        function obj = RWCoseg(L, seeds, lambda, H)
            obj = obj@QPmv(0);
            obj = obj@BoxConstraints(0, 0);
            
            % Get total number of pixels, and ranges into of common indexing
            % scheme for each image.
            obj.nimages = length(L);
            nimages = obj.nimages;
            obj.start = zeros(nimages, 1);
            obj.stop = zeros(nimages, 1);
            obj.npixels = zeros(nimages, 1);
            
            obj.start(1) = 1;
            for i=1:nimages
                obj.npixels(i) = size(L{i}, 1);
                obj.stop(i) = obj.start(i) + obj.npixels(i) - 1;

                if i == nimages
                    n = obj.stop(i);
                else
                    obj.start(i+1) = obj.stop(i) + 1;
                end
            end
            
            
            % We repurpose the box constraints as equality constraints
            % on seeds.  Otherwise, just keep them in [0,1]
            obj.u = ones(n, 1);
            obj.l = zeros(n, 1);
            catseeds = obj.pack(seeds);
            
            obj.u(catseeds == -1) = 0;
            obj.l(catseeds == 1) = 1;
            
            % Set properties for basic image pieces
            obj.seeds = seeds;
            obj.L = L;
            obj.lambda = lambda;
            obj.H = H;
            obj.c = zeros(n, 1);
        end
        
        function rwobj = rwterm(obj, x)
            nimages = length(obj.npixels);
            xi = obj.unpack(x);
            
            rwobj = 0;
            for i=1:nimages
                rwobj = rwobj + xi{i}' * obj.L{i} * xi{i};
            end
            rwobj = rwobj / 2;
        end
        
        function Qx = mv(obj, x)   
            % Calculate the product Q * x
            nimages = length(obj.npixels);
            %n = obj.stop(nimages);
            
            if nimages > 2
                fprintf(1, 'This solver does not support > 2 images.');
                return;
            end
            
            L = obj.L;
            lambda = obj.lambda;
            H = obj.H;
            
            % Split x vector into per-image vectors
            xi = obj.unpack(x);
            
            % Find histograms
            h = cell(nimages, 1);
            for i=1:nimages
                h{i} = H{i} * xi{i};
            end
            
            % Calculate the product
            Qx = cell(2, 1);
            
            Qx{1} = L{1} * xi{1};
            Qx{1} = Qx{1} + lambda * H{1}' * h{1};
            Qx{1} = Qx{1} - lambda * H{1}' * h{2};
            
            Qx{2} = L{2} * xi{2};
            Qx{2} = Qx{2} + lambda * H{2}' * h{2};
            Qx{2} = Qx{2} - lambda * H{2}' * h{1};
            
            % Output concatenation
            Qx = 2*obj.pack(Qx);
        end
        
        function x0 = initial(obj)
            % Perform prior-less RW segmentation, useful as an initial
            % point.
            
            nimages = length(obj.npixels);
            x0n = cell(nimages, 1);
            
            for i=1:nimages
                % Retrieve per-image parameters
                L = obj.L{i};
                seeds = obj.seeds{i};
            
                n = size(L, 1);
                
                % Solve linear system
                x0 = zeros(n, 1);
                x0(seeds == 1) = 1;
                x0(seeds == -1) = 0;
                
                LU = L(seeds == 0, seeds == 0);
                B = L(seeds == 0, seeds ~= 0);
                ms = (seeds(seeds ~= 0) + 1) / 2;
                x0(seeds == 0) = LU \ (-B  * ms);
                
                x0n{i} = x0;
            end
            
            % Output concatenation
            x0 = obj.pack(x0n);
        end
        
        function [x status] = gpu_solve(obj, x0)
            opts.max_iters = 1e4;
            opts.grad_norm = 1e-4;
            opts.print_interval = 100;
            
            % Calculate Jacobi preconditioner
            nimages = length(obj.stop);
            npix = obj.stop(nimages);
            
            if nimages > 2
                fprintf(1, 'This solver does not support > 2 images.');
                return;
            end
            
            v = zeros(npix, 1);
            for i=1:nimages
                v(obj.start(i):obj.stop(i)) = diag(obj.L{i});
            end
            v = v + 1;
            
            i = 1:npix;
            j = i;
            P = sparse(i,j,v,npix,npix);
            Pinv = sparse(i,j,1./v,npix,npix);
            
            % Discard, doesn't work
            Pinv = speye(npix);
            
            [x status] = rwcoseg(obj.L, obj.H, Pinv, Pinv*obj.l, Pinv*obj.u, ...
                                 obj.lambda, x0, opts);
            %x = Pinv * y;
            
            %status.objective = obj.objective(x);
            %status.grad_norm = norm(obj.gradient(x));
        end
        
        function [x status] = cplex_solve(obj, x0)
            fprintf('Setting up CPLEX problem...\n');
            cplex = Cplex('seg');
            
            % Force onto one thread for comparison
            %cplex.Param.threads.Cur = 1;
            
            % Retrieve properties
            nimages = obj.nimages;
            n = obj.stop(nimages);
            m = size(obj.H{1}, 1);
            
            % Add segmentation cols with box + seed constraints
            cplex.addCols(zeros(n, 1), [], obj.l, obj.u);
    
            % Add histogram columns
            cplex.addCols(zeros(m*nimages, 1), [], repmat(-inf, m*nimages, 1), repmat(inf, m*nimages, 1));
    
            % Add histogram model columns
            cplex.addCols(zeros(m, 1), [], repmat(-inf, m, 1), repmat(inf, m, 1));
            
            % Add quadratic coefficients
            Q = cell(2*nimages + 1, 1);
            for i=1:nimages
                Q{i} = obj.L{i}; % L
                Q{nimages + i} = obj.lambda * speye(m); % histograms
            end
            Q{2*nimages + 1} = sparse([], [], [], m, m);
            cplex.Model.Q = spblkdiag(Q);
    
            % Add H_i x_i - hdiff_i - hbar = 0
            b = zeros(m, 1);
            for i=1:nimages
                % H_i x_i
                [Ai Aj Av] = find(obj.H{i});
                Aj = Aj + obj.start(i);
                
                % -hdiff_i
                Ai = [Ai; (1:m)'];
                Aj = [Aj; ((n + m*(i-1) + 1):(n + m*i))'];
                Av = [Av; -ones(m, 1)];
                
                % -hbar
                Ai = [Ai; (1:m)'];
                Aj = [Aj; ((n + m*nimages + 1):(n + m * (nimages + 1)))'];
                Av = [Av; -ones(m, 1)];
                
                
                A = sparse(Ai, Aj, Av, m, n + m * (nimages + 1));
                cplex.addRows(b, A, b);
            end
    
            fprintf('Solving...\n');
            cplex.Solution.x(1:n) = x0;
            cplex.solve();
            x = cplex.Solution.x(1:n);
            hdiff = cell(nimages, 1);
            h = cell(nimages, 1);
            hbar = cplex.Solution.x(end-m+1:end);
            for i=1:nimages
                hdiff{i} = cplex.Solution.x((n + m*(i-1) + 1):(n + m*i));
                h{i} = hbar + hdiff{i};
            end
            fprintf('   objective = %g\n', cplex.Solution.objval);
    
            histterm = obj.lambda * hdiff{i}' * hdiff{i} / 2;
            rwterm = cplex.Solution.objval - histterm;
            best_rwterm = obj.rwterm(obj.initial());
            
            fprintf('   hist term = %g\n', histterm);
            fprintf('     rw term = %g\n', rwterm);
            fprintf('best rw term = %g\n', best_rwterm);
            fprintf('           - = %g\n', rwterm - best_rwterm);
            
            status.objval = cplex.Solution.objval;
            
            clear cplex;
        end
        
        function [x status] = sequence_cplex_solve(obj, x0)
            fprintf('Setting up sequence-comparison CPLEX problem...\n');
            cplex = Cplex('seg');
            
            % Retrieve properties
            nimages = obj.nimages;
            n = obj.stop(nimages);
            m = size(obj.H{1}, 1);
            
            % Add segmentation cols with box + seed constraints
            cplex.addCols(zeros(n, 1), [], obj.l, obj.u);
            
            % Add histogram columns
            cplex.addCols(zeros(m*nimages, 1), [], repmat(-inf, m*nimages, 1), repmat(inf, m*nimages, 1));
            
            % Add quadratic coefficients
            Q = cell(2*nimages, 2*nimages);
            for i=1:nimages
                Q{i, i} = obj.L{i};
                
                if i == 1 || i == nimages
                    Q{nimages + i, nimages + i} = obj.lambda * speye(m);
                else
                    Q{nimages + i, nimages + i} = 2 * obj.lambda * speye(m);
                end
                
                if i ~= 1
                    Q{nimages + i, nimages + i - 1} = -obj.lambda * speye(m);
                    Q{nimages + i - 1, nimages + i} = -obj.lambda * speye(m);
                end
            end
            cplex.Model.Q = spcatcells(Q);
            
            % Add H_i x_i - h_i = 0
            b = zeros(m, 1);
            for i=1:nimages
                % H_i x_i
                [Ai Aj Av] = find(obj.H{i});
                Aj = Aj + obj.start(i);
                
                % -h_i
                Ai = [Ai; (1:m)'];
                Aj = [Aj; ((n + m*(i-1) + 1):(n + m*i))'];
                Av = [Av; -ones(m, 1)];
                
                A = sparse(Ai, Aj, Av, m, n + m * nimages);
                cplex.addRows(b, A, b);
            end
    
            fprintf('Solving...\n');
            cplex.Solution.x(1:n) = x0;
            cplex.solve();
            x = cplex.Solution.x(1:n);
            
            %hdiff = cell(nimages, 1);
            %h = cell(nimages, 1);
            %hbar = cplex.Solution.x(end-m+1:end);
            %for i=1:nimages
            %    hdiff{i} = cplex.Solution.x((n + m*(i-1) + 1):(n + m*i));
            %    h{i} = hbar + hdiff{i};
            %end
            %fprintf('   objective = %g\n', cplex.Solution.objval);
    
            %histterm = obj.lambda * hdiff{i}' * hdiff{i} / 2;
            %rwterm = cplex.Solution.objval - histterm;
            %best_rwterm = obj.rwterm(obj.initial());
            
            %fprintf('   hist term = %g\n', histterm);
            %fprintf('     rw term = %g\n', rwterm);
            %fprintf('best rw term = %g\n', best_rwterm);
            %fprintf('           - = %g\n', rwterm - best_rwterm);
            
            status.objval = cplex.Solution.objval;
            
            clear cplex;
        end
    end
    
end