#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <memory.h>

#include "gonzalez85.h"

#ifndef INFINITY
#define INFINITY (1.0/0.0)
#endif // #ifndef INFINITY

static double geom_distance(size_t ndims, const double *p, const double *q)
{
  size_t i;
  double dist, diff;

  dist = 0;
  for(i = 0; i < ndims; i++) {
    diff = p[i] - q[i];
    dist += diff * diff;
  }
  dist = sqrt(dist);

  return dist;
}

static void geom_vector_scale(size_t ndims, double *p, double s)
{
  size_t i;

  for(i = 0; i < ndims; i++) {
    p[i] *= s;
  }
}

/**
 * A vector all of 1's to serve as dummy weights.
 *
 * @note Caller must free result.
 */
double *ones_vector(size_t n)
{
  double *v;
  size_t i;

  v = malloc(sizeof(double) * n);
  for(i = 0; i < n; i++) {
    v[i] = 1;
  }

  return v;
}

/**
 * Find the k-means centers given a partition.
 */
static void cluster_centers(
  size_t n, size_t ndims, const double *p, const double *w,
  size_t k, const int *part, double *C)
{
  size_t i, j, d;
  double obj, dist, *S;

  double *ones;
  if(w == NULL) {
    ones = ones_vector(n);
    cluster_centers(n, ndims, p, ones, k, part, C);
    free(ones);
    return;
  }

  S = (double *)malloc(sizeof(double) * k);

  // Find centers
  for(j = 0; j < k * ndims; j++) {
    C[j] = 0;
  }
  for(j = 0; j < k; j++) {
    S[j] = 0;
  }

  for(i = 0; i < n; i++) {
    j = (size_t)part[i];
    for(d = 0; d < ndims; d++) {
      C[ndims*j + d] += w[i] * p[ndims*i + d];
    }
    S[j] += w[i];
  }

  for(j = 0; j < k; j++) {
    geom_vector_scale(ndims, C + ndims*j, 1/S[j]);
  }

  free(S);
}

static size_t gonzalez85_real(
  size_t n, size_t ndims, const double *p,
  size_t k, double mindist, double *C, int *part)
{
  size_t i, j, l, vi;
  const double **head;
  double h, *dist, d;
  double *ones;

  head = (const double **)malloc(sizeof(double *) * k);
  ones = (double *)malloc(sizeof(double) * n * 2);
  dist = ones + n;

  // Start everything in first cluster with first point at head
  head[0] = p;
  for(i = 0; i < n; i++) {
    part[i] = 0;
    ones[i] = 1.0;
    dist[i] = geom_distance(ndims, p, p + ndims*i);
  }

  for(l = 1; l < k; l++) {
    // Find most distant point from its center
    h = -INFINITY;
    for(i = 1; i < n; i++) {
      if(dist[i] > h) {
        h = dist[i];
        head[l] = p + ndims*i;
      }
    }

    if(h < mindist) {
      k = l;
      break;
    }

    // Move points into new cluster
    for(i = 1; i < n; i++) {
      d = geom_distance(ndims, head[l], p + ndims * i);

      if(d < dist[i]) {
        dist[i] = d;
        part[i] = l;
      }
    }
  }

  // Calculate centers
  cluster_centers(n, ndims, p, ones, k, part, C);

  // Cleanup
  free(head);

  return k;
}

size_t gonzalez85(
  size_t n, size_t ndims, const double *p,
  size_t k, double mindist,
  double *C, int *part)
{
  size_t ret;

  // Ensure C is non-null
  if(C == NULL) {
    C = (double *)malloc(n * ndims * sizeof(double));
    ret = gonzalez85(n, ndims, p, k, mindist, C, part);
    free(C);
    return ret;
  }

  // Ensure part is non-null
  if(part == NULL) {
    part = (int *)malloc(n * sizeof(int));
    ret = gonzalez85(n, ndims, p, k, mindist, C, part);
    free(part);
    return ret;
  }

  return gonzalez85_real(n, ndims, p, k, mindist, C, part);
}
