function [seg status] = random_walker_cosegmentation(I, H, seeds, params)
    % [seg status] = random_walker_cosegmentation(I, H, seeds, lambda, beta)
    %
    %   The core method for performing random walker coseg.  This code
    %   sets up the problem, formatting it for the optimization step.
    %   The data is then fed to an instance of the RWCoseg class which
    %   provides the optimization methods.
    %
    %
    
    
    % Parse arguments
    defaults.lambda = 0;
    defaults.beta = 1e2;
    defaults.solver = 'cplex';
    defaults.hist_weights = ones(size(I));
    
    if nargin < 4 || numel(params) == 0
        params = defaults;
    else
        params = optdefaults(params, defaults);
    end
    
    nimages = numel(I);
    %if nimages > 2
    %    fprintf(2, 'Don''t actually handle the > 2 image case yet...\n');
    %    seg = {};
    %    return
    %end

    % Create a large laplacian matrix with separate connected components
    % for each image.
    fprintf('Constructing random walker Laplacians...\n');
    tic; L = random_walker_laplacian(I, params.beta); toc;
    %tic; L = random_walker_laplacian(I, params.beta, 'gaussian'); toc;
    
    % Flattening...
    seeds = flatten_image(seeds);
    
    % Get bin count
    m = 0;
    for i=1:nimages
        Hi = H{i};
        m = max(m, max(Hi(:)));
    end
    
    % Construct histogram assignment matrices
    Hmat = histogram_assignment_matrix(m, flatten_image(H));
    for i=1:nimages
        Hmat{i} = Hmat{i} * params.hist_weights(i);
    end
    
    % Create minimization problem
    prob = RWCoseg(L, seeds, params.lambda, Hmat);
    
    % Set initial point to independent RW segmentation
    x0 = prob.initial();
    
    % Solve with CPLEX
    if params.lambda == 0 || nimages == 1
        % Trivial case
        x = x0;
    elseif strcmp(params.solver, 'cplex')
        tic;
        [x status] = prob.cplex_solve(x0);
        status.time = toc;
    elseif strcmp(params.solver, 'sequence_cplex')
        tic;
        [x status] = prob.sequence_cplex_solve(x0);
        status.time = toc;
    else
        tic;
        [x status] = prob.gpu_solve(x0);
        status.time = toc;
        
        %[truth cplex_status] = prob.cplex_solve(x0);
        
        fprintf('Objective Improvement: %e\n', double(prob.objective(truth) - prob.objective(x)));
        misclassified = sum(xor(x > 1/2, truth > 1/2));
        fprintf('# Misclassified: %d (%.1f%%)\n', misclassified, misclassified/length(x) * 100); 
    end
        
    % Unpack into cell array
    seg = prob.unpack(x);
    
    for i=1:nimages
        S = [size(I{i}, 1), size(I{i}, 2)]; % TODO: will need to change for ND
        seg{i} = reshape(seg{i}, S);
    end
end