function L = random_walker_laplacian(I, beta, type, seeds)
    % L = random_walker_laplacian(I, beta)
    %
    %   Generate the laplacian matrix for Random Walker segmentation.
    %
    %    I: An image
    % beta: Multiplier which determines the range of weights.
    % type: One of
    
    if nargin < 3
        type = 'gaussian';
    end

    if iscell(I)
        % Apply to each element of a cell array
        L = cell(size(I));
        n = numel(I);
        
        for i=1:n
            if strcmp(type, 'histogram')
            	L{i} = random_walker_laplacian(I{i}, beta, type, seeds{i});
            else
            	L{i} = random_walker_laplacian(I{i}, beta, type);
            end
        end
        
        return;
    end
        
    % Interpret image size
    X = size(I, 1);
    Y = size(I, 2);
    nchannels = size(I, 3);

    I = im2double(I);

    % Generate lattice graph
    [points edges] = lattice(X, Y);

    % Find edge weights
    switch type
        case 'gaussian'
            I = reshape(I, X*Y, nchannels);
            weights = makeweights(edges, I, beta);
            
        case 'histogram'
            % Construct histogram weights from Grady + Jolly MICCAI 2008
            
            % Get I and fg nodes as coordinates
            I = reshape(I, X*Y, 3);
            fg = I(reshape(seeds, X*Y, 1) > 0, :);
            nfg = size(fg, 1);
            
            % Sample no more than 512 seeds
            if nfg > 512
                nfg = 512;
                fg = fg(randsample(nfg, 512), :);
            end
            
            % Get distances
            normI = sum(I.^2, 2);
            normfg = sum(fg.^2, 2);
            dots = I * fg';
            
            dists = repmat(normI, [1 nfg]) + repmat(normfg, [1 X*Y])' - 2*dots;
            
            % Construct histogram
            H = 1/nfg * sum(exp(-4*dists), 2);
            
            figure; imagesc(reshape(H, X, Y));
            
            % Gaussian weights over histogram
            weights = makeweights(edges, H, beta);
    end

    L = laplacian(edges, weights);        
end